import os
from contextlib import contextmanager
from datetime import datetime
from typing import TextIO, ContextManager, Any, Dict

import yaml
import csv


class DAGsHubLogger:
    """
    A plain Python logger for your metrics and hyperparameters.
    The saved file format is plain and open - CSV for metrics files, YAML for hyperparameters.
    You can use this logger manually, or use one of our integrations with high-level libraries
    like Keras or pytorch-lightning.
    """

    metrics_file: TextIO
    metrics_csv_writer: Any
    hparams: Dict[str, Any]

    def __init__(self,
                 metrics_path: str = 'metrics.csv',
                 should_log_metrics: bool = True,
                 hparams_path: str = 'params.yml',
                 should_log_hparams: bool = True,
                 should_make_dirs: bool = True,
                 eager_logging: bool = True,
                 ):
        """
        :param metrics_path: Where to save the single metrics CSV file.
        :param should_log_metrics: Whether to log metrics at all. Should probably always be True.
        :param hparams_path: Where to save the single hyperparameter YAML file.
        :param should_log_hparams: Whether to log hyperparameters to a file.
            Should be False if you want to work with hyperparameters in a dependency file,
            rather than specifying them using command line arguments.
        :param should_make_dirs: If true, the directory structure required by metrics_path and hparams_path
            will be created. Has no effect if the directory structure already exists.
        :param eager_logging: If true, the logged metrics and hyperparams will be saved to file immediately.
            If false, the logger will wait until save() is called, and until then, will hold metrics and hyperparams
            in memory. Watch out not to run out of memory!
        """
        self.metrics_path = metrics_path
        self.should_log_metrics = should_log_metrics
        self.hparams_path = hparams_path
        self.should_log_hparams = should_log_hparams
        self.should_make_dirs = should_make_dirs
        self.eager_logging = eager_logging

        self.unsaved_metrics = []
        self.hparams = {}

        if eager_logging:
            self.save()

    def log_metrics(self, metrics: Dict[str, Any] = None, step_num: int = 1, **kwargs):
        if self.should_log_metrics:
            copy_of_metrics = dict(metrics or {})
            copy_of_metrics.update(kwargs)
            self.unsaved_metrics.append((copy_of_metrics, self.epoch_milisec(), step_num))
            if self.eager_logging:
                self.save_metrics()

    @staticmethod
    def epoch_milisec():
        microsec_timestamp = datetime.now().timestamp()
        return int(microsec_timestamp * 1000)

    def log_hyperparams(self, params: Dict[str, Any] = None, **kwargs):
        if self.should_log_hparams:
            self.hparams.update(self.normalize_dictionary_values(params or {}))
            self.hparams.update(self.normalize_dictionary_values(kwargs or {}))
            if self.eager_logging:
                self.save_hparams()

    def save(self):
        self.save_metrics()
        self.save_hparams()

    def save_metrics(self):
        if self.should_log_metrics:
            if not hasattr(self, 'metrics_file'):
                self.init_metrics_file()

            for metrics, timestamp, step_num in self.unsaved_metrics:
                for name, value in metrics.items():
                    self.metrics_csv_writer.writerow([name, value, timestamp, step_num])
            self.unsaved_metrics = []

    def save_hparams(self):
        if self.should_log_hparams:
            self.ensure_dir(self.hparams_path)
            with open(self.hparams_path, 'w') as f:
                yaml.safe_dump(self.hparams, f)

    @staticmethod
    def normalize_dictionary_values(dictionary):
        def normalize_dict_deep(dictionary):
            if dictionary is None:
                return None

            new_dict = {}
            for key, value in dictionary.items():
                if isinstance(value, dict):
                    new_dict[key] = normalize_dict_deep(value)
                    continue
                new_dict[key] = value if value is None or type(value) in [int, float, bool] else str(value)

            return new_dict

        return normalize_dict_deep(dictionary)

    def init_metrics_file(self):
        self.ensure_dir(self.metrics_path)
        self.metrics_file = open(self.metrics_path, 'w', newline='')  # newline='' required for csv writer
        self.metrics_file.write("Name,Value,Timestamp,Step\n")
        self.metrics_csv_writer = csv.writer(self.metrics_file, quoting=csv.QUOTE_NONNUMERIC)

    def ensure_dir(self, filename: str):
        if self.should_make_dirs:
            dirname = os.path.dirname(filename)
            if dirname:
                os.makedirs(dirname, exist_ok=True)

    def close(self):
        if hasattr(self, 'metrics_file'):
            self.metrics_file.close()


@contextmanager
def dagshub_logger(metrics_path: str = 'metrics.csv',
                   should_log_metrics: bool = True,
                   hparams_path: str = 'params.yml',
                   should_log_hparams: bool = True,
                   should_make_dirs: bool = True,
                   eager_logging: bool = True, ) -> ContextManager[DAGsHubLogger]:
    """
    Example usage:

    with dagshub_logger() as logger:
        logger.log_hyperparams( {'lr': 1e-3, 'num_layers': 42} )\n
        logger.log_metrics( {'loss': 3.14, 'val_loss': 6.28} )\n
        logger.log_metrics(acc=0.999)\n


    :param metrics_path: Where to save the single metrics CSV file.
    :param should_log_metrics: Whether to log metrics at all. Should probably always be True.
    :param hparams_path: Where to save the single hyperparameter YAML file.
    :param should_log_hparams: Whether to log hyperparameters to a file.
        Should be False if you want to work with hyperparameters in a dependency file,
        rather than specifying them using command line arguments.
    :param should_make_dirs: If true, the directory structure required by metrics_path and hparams_path
        will be created. Has no effect if the directory structure already exists.
    :param eager_logging: If true, the logged metrics and hyperparams will be saved to file immediately.
            If false, the logger will wait until save() is called, and until then, will hold metrics and hyperparams
            in memory. Watch out not to run out of memory!
    """

    logger = DAGsHubLogger(metrics_path=metrics_path, should_log_metrics=should_log_metrics,
                           hparams_path=hparams_path, should_log_hparams=should_log_hparams,
                           should_make_dirs=should_make_dirs, eager_logging=eager_logging)
    try:
        yield logger
    finally:
        logger.save()
        logger.close()



import logging

from dagshub.common.api.responses import RepoAPIResponse, BranchAPIResponse, CommitAPIResponse, StorageAPIEntry, \
    ContentAPIEntry, StorageContentAPIResult
from dagshub.common.util import multi_urljoin

try:
    from functools import cached_property
except ImportError:
    from cached_property import cached_property

from typing import Optional, Tuple, Any, List

import dacite

import dagshub.auth
from dagshub.common import config

from dagshub.common.helpers import http_request

logger = logging.getLogger("dagshub")


class WrongRepoFormatError(Exception):
    pass


class RepoNotFoundError(Exception):
    pass


class BranchNotFoundError(Exception):
    pass


class CommitNotFoundError(Exception):
    pass


class PathNotFoundError(Exception):
    pass


class RepoAPI:

    def __init__(self, repo: str, host: Optional[str] = None, auth: Optional[Any] = None):
        """
        Class for interacting with the API of a repository

        :param repo (str): repo in the format of "user/repo"
        :param host (str): Optional: url of the dagshub instance to connect to
        :param auth: authentication to use to connect
        """
        self.owner, self.repo_name = self.parse_repo(repo)
        self.host = host if host is not None else config.host

        if auth is None:
            self.auth = dagshub.auth.get_authenticator(host=host)
        else:
            self.auth = auth

    def _http_request(self, method, url, **kwargs):
        if "auth" not in kwargs:
            kwargs["auth"] = self.auth
        return http_request(method, url, **kwargs)

    def get_repo_info(self) -> RepoAPIResponse:
        """
        Get information about the repository
        """
        res = self._http_request("GET", self.repo_api_url)

        if res.status_code == 404:
            raise RepoNotFoundError(f"Repo {self.repo_url} not found")
        elif res.status_code >= 400:
            error_msg = f"Got status code {res.status_code} when getting repository info."
            logger.error(error_msg)
            logger.debug(res.content)
            raise RuntimeError(error_msg)
        return dacite.from_dict(RepoAPIResponse, res.json())

    def get_branch_info(self, branch: str) -> BranchAPIResponse:
        """
        Get information about specified branch
        """
        res = self._http_request("GET", self.branch_url(branch))

        if res.status_code == 404:
            raise BranchNotFoundError(f"Branch {branch} not found in repo {self.repo_url}")
        elif res.status_code >= 400:
            error_msg = f"Got status code {res.status_code} when getting branch."
            logger.error(error_msg)
            logger.debug(res.content)
            raise RuntimeError(error_msg)

        return dacite.from_dict(BranchAPIResponse, res.json())

    def get_commit_info(self, sha: str) -> CommitAPIResponse:
        """
        Get information about a specific commit
        """
        res = self._http_request("GET", self.commit_url(sha))

        if res.status_code == 404:
            raise BranchNotFoundError(f"Commit {sha} not found in repo {self.repo_url}")
        elif res.status_code >= 400:
            error_msg = f"Got status code {res.status_code} when getting commit."
            logger.error(error_msg)
            logger.debug(res.content)
            raise RuntimeError(error_msg)

        return dacite.from_dict(CommitAPIResponse, res.json()["commit"])

    def get_connected_storages(self) -> List[StorageAPIEntry]:
        """
        Get storages that are connected to the repository
        """
        res = self._http_request("GET", self.storage_api_url())

        if res.status_code == 404:
            raise RepoNotFoundError(f"Repo {self.repo_url} not found")
        elif res.status_code >= 400:
            error_msg = f"Got status code {res.status_code} when getting repository info."
            logger.error(error_msg)
            logger.debug(res.content)
            raise RuntimeError(error_msg)

        return [dacite.from_dict(StorageAPIEntry, storage_entry) for storage_entry in res.json()]

    def list_path(self, path: str, revision: Optional[str] = None, include_size: bool = False) -> List[ContentAPIEntry]:
        """
        Get listing of everything in the specified path
        """
        params = {
            "include_size": include_size
        }
        res = self._http_request("GET", self.content_api_url(path, revision), params=params)

        if res.status_code == 404:
            raise PathNotFoundError(f"Path {path} not found")
        elif res.status_code >= 400:
            error_msg = f"Got status code {res.status_code} when listing path {path}"
            logger.error(error_msg)
            logger.debug(res.content)
            raise RuntimeError(error_msg)

        content = res.json()
        if type(content) == dict:
            content = [content]
        return [dacite.from_dict(ContentAPIEntry, entry) for entry in content]

    def list_storage_path(self, path: str, include_size: bool = False) -> List[ContentAPIEntry]:
        """
        Get listing of everything in the specified storage path
        Path format: <scheme>/<bucket-name>/<path>
        Example: s3/my-bucket/prefix/path/to/file
        """
        params = {
            "include_size": include_size,
            "paging": True
        }

        url = self.storage_content_api_url(path)
        has_next_page = True

        def _get():
            res = self._http_request("GET", url, params=params)

            if res.status_code == 404:
                raise PathNotFoundError(f"Path {path} not found")
            elif res.status_code >= 400:
                error_msg = f"Got status code {res.status_code} when listing path {path}"
                logger.error(error_msg)
                logger.debug(res.content)
                raise RuntimeError(error_msg)

            return dacite.from_dict(StorageContentAPIResult, res.json())

        entries = []

        while has_next_page:
            has_next_page = False
            resp = _get()
            entries += resp.entries
            if resp.next_token is not None:
                has_next_page = True
                params["from_token"] = resp.next_token

        return entries

    def get_file(self, path: str, revision: Optional[str] = None) -> bytes:
        """
        Download file from repo
        """
        res = self._http_request("GET", self.raw_api_url(path, revision))
        if res.status_code == 404:
            raise PathNotFoundError(f"Path {path} not found")
        elif res.status_code >= 400:
            error_msg = f"Got status code {res.status_code} when getting file {path}"
            logger.error(error_msg)
            logger.debug(res.content)
            raise RuntimeError(error_msg)
        return res.content

    def get_storage_file(self, path: str) -> bytes:
        """
        Download file from storage
        Path format: <scheme>/<bucket-name>/<path>
        Example: s3/my-bucket/prefix/path/to/file
        """
        res = self._http_request("GET", self.storage_raw_api_url(path))
        if res.status_code == 404:
            raise PathNotFoundError(f"Path {path} not found")
        elif res.status_code >= 400:
            error_msg = f"Got status code {res.status_code} when getting file {path}"
            logger.error(error_msg)
            logger.debug(res.content)
            raise RuntimeError(error_msg)
        return res.content

    @cached_property
    def default_branch(self) -> str:
        return self.get_repo_info().default_branch

    @cached_property
    def id(self) -> int:
        return self.get_repo_info().id

    @cached_property
    def is_mirror(self) -> bool:
        return self.get_repo_info().mirror

    @property
    def full_name(self) -> str:
        return f"{self.owner}/{self.repo_name}"

    def last_commit(self, branch: Optional[str] = None) -> CommitAPIResponse:
        if branch is None:
            branch = self.default_branch
        return self.get_branch_info(branch).commit

    def last_commit_sha(self, branch: Optional[str] = None) -> str:
        return self.last_commit(branch).id

    @cached_property
    def repo_api_url(self) -> str:
        """
        Base URL for making all API request for the repos.
        Format: https://dagshub.com/api/v1/repos/user/repo
        """
        return multi_urljoin(
            self.host,
            "api/v1/repos",
            self.owner,
            self.repo_name,
        )

    @cached_property
    def repo_url(self) -> str:
        """
        URL of the repo on DagsHub
        Format: https://dagshub.com/user/repo
        """
        return multi_urljoin(
            self.host,
            self.owner,
            self.repo_name
        )

    def branch_url(self, branch) -> str:
        """
        URL of a branch on the repo
        Format: https://dasghub.com/api/v1/repos/user/repo/branches/branch
        """
        return multi_urljoin(
            self.repo_api_url,
            "branches",
            branch
        )

    @cached_property
    def data_engine_url(self) -> str:
        """
        URL of data engine
        """
        return multi_urljoin(
            self.repo_api_url,
            "data-engine"
        )

    @cached_property
    def annotations_url(self) -> str:
        return multi_urljoin(
            self.repo_api_url,
            "annotations"
        )

    def commit_url(self, sha) -> str:
        """
        URL of a commit in the repo
        Format: https://dagshub.com/api/v1/repos/user/repo/commits/sha
        """
        return multi_urljoin(
            self.repo_api_url,
            "commits",
            sha
        )

    def content_api_url(self, path: str, revision: Optional[str] = None) -> str:
        """
        URL for Content API access
        Format: https://dagshub.com/api/v1/repos/user/repo/content/revision/path
        """
        if revision is None:
            revision = self.default_branch
        return multi_urljoin(
            self.repo_api_url,
            "content",
            revision,
            path
        )

    def raw_api_url(self, path: str, revision: Optional[str] = None) -> str:
        """
        URL for Raw Content API access
        Format: https://dagshub.com/api/v1/repos/user/repo/raw/revision/path
        """
        if revision is None:
            revision = self.default_branch
        return multi_urljoin(
            self.repo_api_url,
            "raw",
            revision,
            path
        )

    def storage_content_api_url(self, path: str) -> str:
        """
        URL for Storage Content API access
        path example: s3/bucket-name/path/in/bucket
        Format: https://dagshub.com/api/v1/repos/user/repo/storage/content/path
        """
        return multi_urljoin(
            self.repo_api_url,
            "storage/content",
            path
        )

    def storage_raw_api_url(self, path: str) -> str:
        """
        URL for Storage Raw Content API access
        path example: s3/bucket-name/path/in/bucket
        Format: https://dagshub.com/api/v1/repos/user/repo/storage/raw/path
        """
        return multi_urljoin(
            self.repo_api_url,
            "storage/raw",
            path
        )

    def storage_api_url(self) -> str:
        """
        URL for getting connected storages
        Format: https://dagshub.com/api/v1/repos/user/repo/storage
        """
        return multi_urljoin(self.repo_api_url, "storage")

    @staticmethod
    def parse_repo(repo: str) -> Tuple[str, str]:
        repo = repo.strip("/")
        parts = repo.split("/")
        if len(parts) != 2:
            raise WrongRepoFormatError("repo needs to be in the format <repo-owner>/<repo-name>")
        return parts[0], parts[1]



import logging
import urllib
from os.path import ismount
from pathlib import Path

import httpx

from dagshub.common import config, rich_console

default_logger = logging.getLogger("dagshub")


def get_default_branch(owner, reponame, auth, host=config.host):
    """
    The get_default_branch function returns the default branch of a given repository.

    :param owner(str): Specify the owner of the repository
    :param reponame (str): Specify the name of the repository
    :param auth: Authentication object or a (username, password) tuple
    :param host (str): Specify the host to be used
    :return: The default branch of the given repository
    """
    res = http_request("GET", urllib.parse.urljoin(host, config.REPO_INFO_URL.format(
        owner=owner,
        reponame=reponame,
    )), auth=auth)
    return res.json().get('default_branch')


def http_request(method, url, **kwargs):
    mixin_args = {
        "timeout": config.http_timeout,
        "follow_redirects": True
    }
    # Set only if it's not set previously
    for arg in mixin_args:
        if arg not in kwargs:
            kwargs[arg] = mixin_args[arg]
    # Add the config headers to the headers being sent out
    headers = kwargs.get("headers", {})
    headers.update(config.requests_headers)
    kwargs["headers"] = headers
    return httpx.request(method, url, **kwargs)


def get_project_root(root):
    while not (root / '.git').is_dir():
        if ismount(root):
            raise ValueError(f"No git project found! (stopped at mountpoint {root}). \
                               Please run this command in a git repository.")
        root = root / '..'
    return Path(root)


def sizeof_fmt(num, suffix="B"):
    """
    Shoutout to https://stackoverflow.com/a/1094933
    """
    for unit in ["", "Ki", "Mi", "Gi", "Ti", "Pi", "Ei", "Zi"]:
        if abs(num) < 1024.0:
            return f"{num:3.1f}{unit}{suffix}"
        num /= 1024.0
    return f"{num:.1f}Yi{suffix}"


def prompt_user(prompt, default=False) -> bool:
    """
    Prompt the user for input.
    Asks them the provided prompt + Are you sure [letters].
    Default value can be specified with the `default` args.
    """
    prompt_letters = "[y/(N)]" if not default else "[(Y)/n]"
    prompt += f"\nAre you sure {prompt_letters}: "

    prompt_response = input(prompt).lower()
    if prompt_response not in ["y", "n"]:
        return default
    return prompt_response == "y"


def log_message(msg, logger=None):
    """
    Logs message to the info of the logger + prints, unless the printing was suppressed
    """
    if not config.quiet:
        rich_console.print(msg)
    logger = logger or default_logger
    logger.info(msg)



import logging

from starlette.applications import Starlette
from starlette.middleware import Middleware
from starlette.middleware.cors import CORSMiddleware
from starlette.responses import JSONResponse
from starlette.routing import Route

from dagshub.data_engine.voxel_plugin_server.routes.datasource import get_fields, update_metadata, refresh_dataset
from dagshub.data_engine.voxel_plugin_server.routes.annotation import to_annotate
from dagshub.data_engine.voxel_plugin_server.routes.voxel import save_dataset

logger = logging.getLogger(__name__)


async def homepage(request):
    return JSONResponse({"Hello": "from the dagshub voxel plugin server"})


app = Starlette(debug=True,
                middleware=[
                    Middleware(
                        CORSMiddleware,
                        allow_origins=["*"],
                        allow_methods=["*"],
                    ),
                ],
                routes=[
                    Route("/", homepage),
                    Route("/labelstudio/", to_annotate, methods=["POST"]),
                    Route("/dataset/save", save_dataset, methods=["POST"]),
                    Route("/dataset/refresh", refresh_dataset),
                    Route("/datasource/fields", get_fields),
                    Route("/datasource/update_metadata", update_metadata, methods=["POST"]),
                ])



# Vendored library: youyo/gql-query-builder
# LICENSE: MIT
# Repo: https://github.com/youyo/gql-query-builder/

# Vendored due to problems in the installation process
# (TODO: check if issue https://github.com/youyo/gql-query-builder/issues/8 was fixed )


# coding: utf-8


from typing import Dict, List, Union


class GqlQuery:
    def __init__(self) -> None:
        self.object: str = ''
        self.return_field: str = ''
        self.query_field: str = ''
        self.operation_field: str = ''
        self.fragment_field: str = ''

    def remove_duplicate_spaces(self, query: str) -> str:
        return " ".join(query.split())

    def fields(self, fields: List, name: str = '', condition_expression: str = ''):
        query = '{ ' + " ".join(fields) + ' }'
        if name != '':
            if condition_expression != '':
                query = f'{name} {condition_expression} {query}'
            else:
                query = f'{name} {query}'
        self.return_field = query
        return self

    @staticmethod
    def build_input(input: Dict[str, Union[str, int]], initial_str: str):
        inputs: List[str] = []

        final_str = initial_str

        if input != {}:
            key = list(input.keys())[0]
            nested_keys = list()

            while isinstance(input[key], dict):
                nested_keys.append(key)
                input = input[key]
                key = list(input.keys())[0]

            for key, value in input.items():
                if nested_keys:
                    inputs.append(f'{key}: "{value}"')  # Nested input won't have double quotes

                else:
                    inputs.append(f'{key}: {value}')

            final_str += '('

            for key in nested_keys:
                final_str = final_str + key + ': {'

            final_str = final_str + ", ".join(inputs)

            for _ in nested_keys:
                final_str += '}'

            final_str += ')'

        return final_str

    def query(self, name: str, alias: str = '', input: Dict[str, Union[str, int]] = {}):
        self.query_field = name
        self.query_field = self.build_input(input, self.query_field)
        if alias != '':
            self.query_field = f'{alias}: {self.query_field}'

        return self

    def operation(self, query_type: str = 'query', name: str = '', input: Dict[str, Union[str, int]] = {},
                  queries: List[str] = []):
        self.operation_field = query_type
        if name != '':
            self.operation_field = f'{self.operation_field} {name}'
            self.operation_field = self.build_input(input, self.operation_field)

        if queries != []:
            self.object = self.operation_field + ' { ' + " ".join(queries) + ' }'

        return self

    def fragment(self, name: str, interface: str):
        self.fragment_field = f'fragment {name} on {interface}'
        return self

    def generate(self) -> str:
        if self.fragment_field != '':
            self.object = f'{self.fragment_field} {self.return_field}'
        else:
            if self.object == '' and self.operation_field == '' and self.query_field == '':
                self.object = self.return_field
            elif self.object == '' and self.operation_field == '':
                self.object = self.query_field + ' ' + self.return_field
            elif self.object == '':
                self.object = self.operation_field + ' { ' + self.query_field + ' ' + self.return_field + ' }'

        return self.remove_duplicate_spaces(self.object)



import errno
import logging
import os
import platform
import sys
from argparse import ArgumentParser
from os import PathLike
from pathlib import Path
from threading import Lock
from typing import Optional
from dagshub.common import rich_console

from .filesystem import SPECIAL_FILE, DagsHubFilesystem

logger = logging.getLogger(__name__)

SPECIAL_FILE_FH = (1 << 64) - 1

fuse_enabled_systems = ["Linux"]
system = platform.system()
if system not in fuse_enabled_systems:
    err_str = f"FUSE mounting isn't supported on {system}.\n" \
              f"Please use install_hooks to access DagsHub hosted files from a python script"
    raise ImportError(err_str)
from fuse import FUSE, FuseOSError, LoggingMixIn, Operations  # noqa


class DagsHubFUSE(LoggingMixIn, Operations):
    def __init__(self,
                 project_root: Optional[PathLike] = None,
                 repo_url: Optional[str] = None,
                 branch: Optional[str] = None,
                 username: Optional[str] = None,
                 password: Optional[str] = None,
                 token: Optional[str] = None):
        # FIXME TODO move autoconfiguration out of FUSE object constructor and to main method
        self.fs = DagsHubFilesystem(project_root=project_root, repo_url=repo_url, branch=branch, username=username,
                                    password=password, token=token)
        logger.debug("__init__")
        self.rwlock = Lock()

    def __call__(self, op, path, *args):
        return super(DagsHubFUSE, self).__call__(op, self.fs.project_root / path[1:], *args)

    def access(self, path, mode):
        logger.debug(f"access - path: {path}, mode:{mode}")
        try:
            self.fs.stat(path)
        except FileNotFoundError:
            return False

    def open(self, path, flags):
        logger.debug(f"open - path: {path}, flags: {flags}")
        if path == Path(self.fs.project_root / SPECIAL_FILE):
            return SPECIAL_FILE_FH
        try:
            self.fs.open(path).close()
        except FileNotFoundError:
            raise FuseOSError(errno.ENOENT)
        logger.debug("finished fs.open")
        return os.open(self.fs._relative_path(path), flags, dir_fd=self.fs.project_root_fd)

    def getattr(self, path, fd=None):
        logger.debug(f"getattr - path:{str(path)}, fd:{fd}")
        try:
            if fd:
                logger.debug("with __stat")
                st = self.fs._DagsHubFilesystem__stat(fd)
            else:
                logger.debug("with fs.stat")
                st = self.fs.stat(path)

            logger.debug(f"st: {st}")
            return {
                key: getattr(st, key)
                for key in (
                    'st_atime',
                    'st_ctime',
                    'st_gid',
                    'st_mode',
                    'st_mtime',
                    # 'st_nlink',
                    'st_size',
                    'st_uid',
                )
            }
        except FileNotFoundError:
            logger.debug("FileNotFound")
            raise FuseOSError(errno.ENOENT)

    def read(self, path, size, offset, fh):
        logger.debug(f"read - path: {path}, offset: {offset}, fh: {fh}")
        if fh == SPECIAL_FILE_FH:
            return self.fs._special_file()[offset:offset + size]
        with self.rwlock:
            os.lseek(fh, offset, 0)
            return os.read(fh, size)

    def readdir(self, path, fh):
        logger.debug(f"readdir - path: {path}, fh: {fh}")
        return ['.', '..'] + self.fs.listdir(path)

    def release(self, path, fh):
        logger.debug(f"release - path: {path}, fh: {fh}")
        if fh != SPECIAL_FILE_FH:
            return os.close(fh)


def mount(debug=False,
          project_root: Optional[PathLike] = None,
          repo_url: Optional[str] = None,
          branch: Optional[str] = None,
          username: Optional[str] = None,
          password: Optional[str] = None,
          token: Optional[str] = None):
    logging.basicConfig(level=logging.DEBUG)
    fuse = DagsHubFUSE(project_root=project_root, repo_url=repo_url, branch=branch, username=username,
                       password=password, token=token)
    rich_console.print(
        f'Mounting DagsHubFUSE filesystem at {fuse.fs.project_root}\n'
        f'Run `cd .` in any existing terminals to utilize mounted FS.')
    FUSE(fuse, str(fuse.fs.project_root), foreground=debug, nonempty=True)
    if not debug:
        os.chdir(os.path.realpath(os.curdir))
    # TODO: Clean unmounting procedure


def main():
    parser = ArgumentParser()
    parser.add_argument('project_root', nargs='?')
    parser.add_argument('--repo_url')
    parser.add_argument('--branch')
    parser.add_argument('--username')
    parser.add_argument('--password')
    parser.add_argument('--debug', action='store_true', default=False)  # default=False, nargs=0)

    args = parser.parse_args()

    if not args.debug:
        # Hide tracebacks of errors, display only error message
        sys.tracebacklimit = 0

    mount(**vars(args))


if __name__ == '__main__':
    main()



import pytest

from dagshub.common.download import download_url_to_bucket_path


@pytest.mark.parametrize("in_url, expected", [
    ("https://dagshub.com/api/v1/repos/kirill/bucket-repo/storage/raw/s3/dagshub-storage/images/047.png",
     ("s3", "dagshub-storage", "images/047.png"))
])
def test_bucket_downloader_path_extraction(in_url, expected):
    actual = download_url_to_bucket_path(in_url)
    assert actual == expected


@pytest.mark.parametrize("in_url", [
    "https://dagshub.com/api/v1/repos/kirill/bucket-repo/raw/main/images/047.png",
    "https://google.com"
])
def test_bucket_download_path_extraction_fail(in_url):
    actual = download_url_to_bucket_path(in_url)
    assert actual is None



from dataclasses import dataclass
from os import PathLike
from pathlib import Path
from typing import Tuple, Optional, Union, List, Dict, Any, Callable, TYPE_CHECKING

from dagshub.common.download import download_files
from dagshub.common.helpers import http_request
from dagshub.data_engine.client.models import MetadataFieldType

if TYPE_CHECKING:
    from dagshub.data_engine.model.datasource import Datasource

_generated_fields: Dict[str, Callable[["Datapoint"], Any]] = {
    "path": lambda dp: dp.path,
    "datapoint_id": lambda dp: dp.datapoint_id,
    "dagshub_download_url": lambda dp: dp.download_url,
}


@dataclass
class Datapoint:
    datapoint_id: int
    path: str
    metadata: Dict[str, Any]
    datasource: "Datasource"

    def __getitem__(self, item):
        gen_field = _generated_fields.get(item)
        if gen_field is not None:
            return gen_field(self)
        return self.metadata[item]

    @property
    def download_url(self):
        return self.datasource.source.raw_path(self)

    @property
    def path_in_repo(self):
        return self.datasource.source.file_path(self)

    @staticmethod
    def from_gql_edge(edge: Dict, datasource: "Datasource") -> "Datapoint":
        res = Datapoint(
            datapoint_id=int(edge["node"]["id"]),
            path=edge["node"]["path"],
            metadata={},
            datasource=datasource,
        )

        float_fields = {f.name for f in datasource.fields if f.valueType == MetadataFieldType.FLOAT}

        for meta_dict in edge["node"]["metadata"]:
            key = meta_dict["key"]
            value = meta_dict["value"]
            if key in float_fields:
                value = float(value)
            res.metadata[key] = value
        return res

    def to_dict(self, metadata_keys: List[str]) -> Dict[str, Any]:
        # Set the autogenerated fields
        res_dict = {k: v(self) for k, v in _generated_fields.items()}
        res_dict.update({key: self.metadata.get(key) for key in metadata_keys})
        return res_dict

    def get_blob(self, column: str, cache_on_disk=True, store_value=False) -> bytes:
        """
        Returns the blob stored in the binary column

        Args:
            column: where to get the blob from
            cache_on_disk: whether to store the downloaded blob on disk.
                If you store the blob on disk, it means that it won't need to be re-downloaded in the future.
                The contents of datapoint[column] will change to be the path of the blob on the disk.
            store_value: whether to store the blob in memory on the field attached to this datapoint,
                which will make its bytes directly retrievable using datapoint[column]
        """
        current_value = self.metadata[column]

        if type(current_value) is bytes:
            # Bytes - it's already there!
            return current_value
        if isinstance(current_value, Path):
            # Path - assume the path exists and is already downloaded,
            #   because it's unlikely that the user has set it themselves
            with current_value.open("rb") as f:
                content = f.read()
            if store_value:
                self.metadata[column] = content
            return content

        elif type(current_value) is str:
            # String - This is probably the hash of the blob, get that from dagshub
            blob_url = self.blob_url(current_value)
            blob_location = self.blob_cache_location / current_value

            # Make sure that the cache location exists
            if cache_on_disk:
                self.blob_cache_location.mkdir(parents=True, exist_ok=True)

            content = _get_blob(blob_url, blob_location, self.datasource.source.repoApi.auth, cache_on_disk, True)
            if type(content) is str:
                raise RuntimeError(f"Error while downloading blob: {content}")

            if store_value:
                self.metadata[column] = content
            elif cache_on_disk:
                self.metadata[column] = blob_location

            return content
        else:
            raise ValueError(f"Can't extract blob metadata from value {current_value} of type {type(current_value)}")

    def download_file(self, target: Optional[Union[PathLike, str]] = None, keep_source_prefix=True,
                      redownload=False) -> PathLike:
        """
        Downloads the datapoint to the target_dir directory
        Args:
            target: Where to download the file (either a directory, or the full path).
                If not specified, then downloads to ~/dagshub/datasets/<user>/<repo>/<ds_id>
            keep_source_prefix: If True, includes the prefix of the datasource in the download path
                Useful for cases where the download path is the root of the repository
            redownload: Whether to redownload a file if it exists on the filesystem already
                NOTE: We don't do any hashsum checks, so if it's possible that the file has been updated, turn it on
        Returns:
            Path to the downloaded file
        """

        target_path = self.datasource.default_dataset_location if target is None else Path(target).expanduser()

        # Check if the specified path looks like a file
        # by checking if there's an extension, or it's an already existing file
        n = target_path.name
        target_is_already_file = (target_path.exists() and target_path.is_file()) or (
            "." in n and not n.startswith("."))

        if not target_is_already_file:
            if keep_source_prefix:
                target_path = target_path / self.path_in_repo
            else:
                target_path = target_path / self.path

        download_files([(self.download_url, target_path)], skip_if_exists=not redownload)
        return target_path

    @property
    def blob_cache_location(self):
        return self.datasource.default_dataset_location / ".metadata_blobs"

    def blob_url(self, sha):
        return self.datasource.source.blob_path(sha)

    def _extract_blob_url_and_path(self, col: str) -> Tuple[Optional[str], Optional[PathLike]]:
        sha = self.metadata.get(col)
        if sha is None or type(sha) is not str:
            return None, None
        return self.blob_url(sha), self.blob_cache_location / sha


def _get_blob(url: Optional[str], cache_path: Optional[Path], auth, cache_on_disk, return_blob) -> Optional[
    Union[Path, str, bytes]]:
    """
    Args:
        url: url to download the blob from
        cache_path: where the cache for the blob is (laods from it if exists, stores there if it doesn't)
        auth: auth to use for getting the blob
        cache_on_disk: whether to store the downloaded blob on disk. If False we also turn off the cache checking
        return_blob: if True returns the blob of the downloaded data, if False returns the path to the file with it
    """
    if url is None:
        return None
    assert cache_path is not None

    if cache_on_disk and cache_path.exists():
        if return_blob:
            with cache_path.open("rb") as f:
                return f.read()
        else:
            return cache_path

    try:
        # TODO: add retries here
        resp = http_request("GET", url, auth=auth)
        if resp.status_code >= 400:
            return f"Error while downloading binary blob (Status code {resp.status_code}): {resp.content.decode()}"
        content = resp.content
    except Exception as e:
        return f"Error while downloading binary blob: {e}"

    if cache_on_disk:
        with cache_path.open("wb") as f:
            f.write(content)

    if return_blob:
        return content
    else:
        return cache_path



import enum
import logging
from dataclasses import dataclass, field
from typing import Any, List, Union, Optional

from dataclasses_json import dataclass_json, config

logger = logging.getLogger(__name__)


@dataclass
class Metadata:
    key: str
    value: Any


autogenerated_columns = {
    "path",
    "datapoint_id",
    "dagshub_download_url",
}


class IntegrationStatus(enum.Enum):
    VALID = "VALID"
    INVALID_CREDENTIALS = "INVALID_CREDENTIALS"
    MISSING = "MISSING"


class PreprocessingStatus(enum.Enum):
    READY = "READY"
    IN_PROGRESS = "IN_PROGRESS"
    AUTO_SCAN_IN_PROGRESS = "AUTO_SCAN_IN_PROGRESS"
    FAILED = "FAILED"


class DatasourceType(enum.Enum):
    BUCKET = "BUCKET"
    REPOSITORY = "REPOSITORY"
    CUSTOM = "CUSTOM"


class MetadataFieldType(enum.Enum):
    BOOLEAN = "BOOLEAN"
    INTEGER = "INTEGER"
    FLOAT = "FLOAT"
    STRING = "STRING"
    BLOB = "BLOB"


class ScanOption(str, enum.Enum):
    FORCE_REGENERATE_AUTO_SCAN_VALUES = "FORCE_REGENERATE_AUTO_SCAN_VALUES"


@dataclass_json
@dataclass
class MetadataFieldSchema:
    name: str
    valueType: MetadataFieldType = field(
        metadata=config(
            encoder=lambda val: val.value
        )
    )
    multiple: bool

    def __repr__(self):
        return f"{self.name} ({self.valueType.value})"


@dataclass
class DatasourceResult:
    id: Union[str, int]
    name: str
    rootUrl: str
    integrationStatus: IntegrationStatus
    preprocessingStatus: PreprocessingStatus
    type: DatasourceType
    metadataFields: Optional[List[MetadataFieldSchema]]


@dataclass
class DatasetResult:
    id: Union[str, int]
    name: str
    datasource: DatasourceResult
    datasetQuery: str



import logging
from typing import Any, Optional, List, Dict, Union, TYPE_CHECKING

import dacite
import gql
import rich.progress
from gql.transport.requests import RequestsHTTPTransport

import dagshub.auth
import dagshub.common.config
from dagshub.common import config
from dagshub.common.analytics import send_analytics_event
from dagshub.common.rich_util import get_rich_progress
from dagshub.data_engine.client.models import DatasourceResult, DatasourceType, IntegrationStatus, \
    PreprocessingStatus, DatasetResult, MetadataFieldType, ScanOption
from dagshub.data_engine.client.gql_mutations import GqlMutations
from dagshub.data_engine.client.gql_queries import GqlQueries
from dagshub.data_engine.model.datasource import Datasource, DatapointMetadataUpdateEntry
from dagshub.data_engine.model.query_result import QueryResult

if TYPE_CHECKING:
    from dagshub.data_engine.datasources import DatasourceState

logger = logging.getLogger(__name__)

dacite_config = dacite.Config(cast=[IntegrationStatus, DatasourceType, PreprocessingStatus, MetadataFieldType])


class DataClient:
    HEAD_QUERY_SIZE = 100
    FULL_LIST_PAGE_SIZE = 5000

    def __init__(self, repo: str):
        # TODO: add project authentication here
        self.repo = repo
        self.host = config.host
        self.client = self._init_client()

    def _init_client(self):
        url = f"{self.host}/api/v1/repos/{self.repo}/data-engine/graphql"
        auth = dagshub.auth.get_authenticator(host=self.host)
        transport = RequestsHTTPTransport(url=url, auth=auth)
        client = gql.Client(transport=transport)
        return client

    def create_datasource(self, ds: "DatasourceState") -> DatasourceResult:
        q = GqlMutations.create_datasource()

        assert ds.name is not None
        assert ds.path is not None
        assert ds.source_type is not None

        params = GqlMutations.create_datasource_params(
            name=ds.name,
            url=ds.path,
            ds_type=ds.source_type
        )
        res = self._exec(q, params)
        return dacite.from_dict(DatasourceResult, res["createDatasource"], config=dacite_config)

    def head(self, datasource: Datasource, size: Optional[int] = None) -> QueryResult:
        if size is None:
            size = self.HEAD_QUERY_SIZE

        progress = get_rich_progress(rich.progress.MofNCompleteColumn())
        total_task = progress.add_task("Downloading metadata...", total=size)

        with progress:
            resp = self._datasource_query(datasource, True, size)
            progress.update(total_task, advance=size, refresh=True)

        return QueryResult.from_gql_query(resp, datasource)

    def sample(self, datasource: Datasource, n: Optional[int], include_metadata: bool) -> QueryResult:
        if n is None:
            return self._get_all(datasource, include_metadata)

        has_next_page = True
        after = None
        res = QueryResult([], datasource)
        left = n

        progress = get_rich_progress(rich.progress.MofNCompleteColumn())
        total_task = progress.add_task("Downloading metadata...", total=left)

        with progress:
            while has_next_page and left > 0:
                take = min(left, self.FULL_LIST_PAGE_SIZE)
                resp = self._datasource_query(datasource, include_metadata, take, after)
                has_next_page = resp["pageInfo"]["hasNextPage"]
                after = resp["pageInfo"]["endCursor"]
                new_entries = QueryResult.from_gql_query(resp, datasource)
                res.entries += new_entries.entries
                left -= take
                progress.update(total_task, advance=len(res.entries), refresh=True)
        return res

    def get_datapoints(self, datasource: Datasource) -> QueryResult:
        return self._get_all(datasource, True)

    def _get_all(self, datasource: Datasource, include_metadata: bool) -> QueryResult:
        has_next_page = True
        after = None
        res = QueryResult([], datasource)
        # TODO: smarter batch sizing. Query a constant size at first
        #       On next queries adjust depending on the amount of metadata columns

        progress = get_rich_progress(rich.progress.MofNCompleteColumn())
        total_task = progress.add_task("Downloading metadata...", total=None)

        with progress:
            while has_next_page:
                resp = self._datasource_query(datasource, include_metadata, self.FULL_LIST_PAGE_SIZE, after)
                has_next_page = resp["pageInfo"]["hasNextPage"]
                after = resp["pageInfo"]["endCursor"]

                new_entries = QueryResult.from_gql_query(resp, datasource)
                res.entries += new_entries.entries
                progress.update(total_task, advance=len(new_entries.entries), refresh=True)
        return res

    def _exec(self, query: str, params: Optional[Dict[str, Any]] = None) -> Dict[str, Any]:
        logger.debug(f"Executing query: {query}")
        if params is not None:
            logger.debug(f"Params: {params}")
        q = gql.gql(query)
        resp = self.client.execute(q, variable_values=params)
        return resp

    def _datasource_query(self, datasource: Datasource, include_metadata: bool, limit: Optional[int] = None,
                          after: Optional[str] = None):

        send_analytics_event("Client_DataEngine_QueryRun", repo=datasource.source.repoApi)

        q = GqlQueries.datasource_query(include_metadata)

        params = GqlQueries.datasource_query_params(
            datasource_id=datasource.source.id,
            query_input=datasource.serialize_gql_query_input(),
            first=limit,
            after=after
        )

        return self._exec(q, params)["datasourceQuery"]

    def update_metadata(self, datasource: Datasource, entries: List[DatapointMetadataUpdateEntry]):
        q = GqlMutations.update_metadata()

        assert datasource.source.id is not None
        assert len(entries) > 0

        params = GqlMutations.update_metadata_params(
            datasource_id=datasource.source.id,
            datapoints=[e.to_dict() for e in entries]
        )

        return self._exec(q, params)

    def get_datasources(self, id: Optional[str], name: Optional[str]) -> List[DatasourceResult]:
        q = GqlQueries.datasource()
        params = GqlQueries.datasource_params(id=id, name=name)

        res = self._exec(q, params)["datasource"]
        if res is None:
            return []
        return [dacite.from_dict(DatasourceResult, val, config=dacite_config)
                for val in res]

    def delete_datasource(self, datasource: Datasource):
        q = GqlMutations.delete_datasource()

        assert datasource.source.id is not None

        params = GqlMutations.delete_datasource_params(datasource_id=datasource.source.id)
        return self._exec(q, params)

    def scan_datasource(self, datasource: Datasource, options: Optional[List[ScanOption]]):
        q = GqlMutations.scan_datasource()

        assert datasource.source.id is not None

        params = GqlMutations.scan_datasource_params(datasource_id=datasource.source.id, options=options)
        return self._exec(q, params)

    def save_dataset(self, datasource: Datasource, name: str):
        q = GqlMutations.save_dataset()

        assert name is not None

        params = GqlMutations.save_dataset_params(datasource_id=datasource.source.id,
                                                  name=name,
                                                  query_input=datasource.serialize_gql_query_input())
        return self._exec(q, params)

    def get_datasets(self, id: Optional[Union[str, int]], name: Optional[str]) -> List[DatasetResult]:
        q = GqlQueries.dataset()
        params = GqlQueries.dataset_params(id=id, name=name)

        res = self._exec(q, params)["dataset"]
        if res is None:
            return []

        return [dacite.from_dict(DatasetResult, val, config=dacite_config)
                for val in res]



import rich.console


def is_inside_notebook():
    try:
        from IPython import get_ipython
        return get_ipython() is not None
    except ModuleNotFoundError:
        return False


def is_inside_colab():
    if not is_inside_notebook():
        return False
    from IPython import get_ipython
    return "google.colab" in get_ipython().extension_manager.loaded


force_terminal = False if is_inside_notebook() else None

rich_console = rich.console.Console(force_terminal=force_terminal)



import rich.progress

from dagshub.common import rich_console, config


def get_rich_progress(*additional_columns, transient=True):
    progress = rich.progress.Progress(rich.progress.SpinnerColumn(), *rich.progress.Progress.get_default_columns(),
                                      *additional_columns,
                                      console=rich_console, transient=transient, disable=config.quiet)
    return progress



import json
import logging
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from dagshub.data_engine.client.models import Datapoint
    import fiftyone as fo

logger = logging.getLogger(__name__)


def add_voxel_annotations(sample: "fo.Sample", datapoint: "Datapoint", *annotation_fields: str):
    from fiftyone import Label
    for field in annotation_fields:
        annotation_val = datapoint.get_blob(field)
        label = Label.from_json(annotation_val.decode())
        sample.add_labels(label, label_field=field)


def add_ls_annotations(sample: "fo.Sample", datapoint: "Datapoint", *annotation_fields: str):
    """
    Adds LabelStudio annotation to the voxel sample.

    Args:
        sample: FiftyOne sample to add the annotations to
        datapoint: Data Engine datapoint to get metadata from
        annotation_fields: fields from which to get annotations
    """
    from fiftyone.utils.labelstudio import import_label_studio_annotation
    from fiftyone import Detections, Detection, Classification, Classifications, Keypoint, Keypoints, Polylines, \
        Polyline
    for field in annotation_fields:
        annotations = datapoint.metadata.get(field)
        if type(annotations) is not bytes:
            return
        ann_dict = json.loads(annotations.decode())
        for ann in ann_dict.get("annotations", {}):
            if "result" not in ann:
                continue
            annotations = []
            for res in ann["result"]:
                try:
                    converted = import_label_studio_annotation(res)
                    annotations.append(converted)
                except Exception:
                    logger.warning(f"Couldn't convert LS annotation {ann} to voxel annotation")

            if len(annotations) == 0:
                continue
            # Group the annotations of a similar type together
            # For now assuming there's no mixing and matching
            ann_type = type(annotations[0])
            if ann_type is Detection:
                labels = Detections()
                for a in annotations:
                    labels.detections.append(a)
                labels = [labels]
            elif ann_type is Classification:
                labels = Classifications()
                for a in annotations:
                    labels.classifications.append(a)
                labels = [labels]
            elif ann_type is Keypoint:
                labels = Keypoints()
                for a in annotations:
                    labels.keypoints.append(a)
                labels = [labels]
            elif ann_type is Polyline:
                labels = Polylines()
                for a in annotations:
                    labels.polylines.append(a)
                labels = [labels]
            else:
                labels = annotations

            for label in labels:
                sample.add_labels(label, label_field=field)



import logging
from dataclasses import dataclass
from typing import Any

import dacite
from starlette.requests import Request
from starlette.responses import JSONResponse

from dagshub.common.analytics import send_analytics_event
from dagshub.data_engine.client.data_client import dacite_config
from dagshub.data_engine.client.models import MetadataFieldSchema
from dagshub.data_engine.voxel_plugin_server.routes.util import error_handler
from dagshub.data_engine.voxel_plugin_server.utils import get_plugin_state

logger = logging.getLogger(__name__)


@error_handler
async def get_fields(request: Request):
    plugin_state = get_plugin_state(request)
    ds = plugin_state.datasource
    return JSONResponse([a.to_json() for a in ds.source.metadata_fields])


@dataclass
class UpdateMetadataRequestData:
    field: MetadataFieldSchema
    value: Any


@error_handler
async def update_metadata(request: Request):
    plugin_state = get_plugin_state(request)
    ds = plugin_state.datasource
    voxel_sess = plugin_state.voxel_session

    send_analytics_event("Client_DataEngine_addEnrichmentsWithVoxel", repo=ds.source.repoApi)

    req_data = dacite.from_dict(UpdateMetadataRequestData, await request.json(), config=dacite_config)
    datapoints = [sample["datapoint_path"] for sample in voxel_sess.selected_view]

    with ds.metadata_context() as ctx:
        ctx.update_metadata(datapoints, {req_data.field.name: req_data.value})

    return JSONResponse("OK")


@error_handler
async def refresh_dataset(request: Request):
    plugin_state = get_plugin_state(request)
    ds = plugin_state.datasource
    voxel_sess = plugin_state.voxel_session
    current_dataset = voxel_sess.dataset

    updated_dataset = ds.to_voxel51_dataset(name=current_dataset.name, force_download=True)
    updated_dataset.save()
    voxel_sess.refresh()

    return JSONResponse("OK")



from tensorflow.keras.callbacks import Callback

from ..logger import DAGsHubLogger as LoggerImpl


class ignore_exceptions:
    # Taken from fastcore code
    """Context manager to ignore exceptions"""

    def __enter__(self):
        pass

    def __exit__(self, *args):
        return True


class DAGsHubLogger(Callback):
    """
    First, install the DAGsHub logger with
    `pip install dagshub`

    Example of Usage:
    ```
    from dagshub.keras import DAGsHubLogger
    from tensorflow.keras import Trainer

    model = ...

    model.fit(
        dataset,
        ...,
        callbacks=[DAGsHubLogger()]
    )
    ```
    """

    def __init__(self,
                 metrics_path: str = 'metrics.csv',
                 should_log_metrics: bool = True,
                 hparams_path: str = 'params.yml',
                 should_log_hparams: bool = True,
                 should_make_dirs: bool = True,
                 ):
        """
        :param metrics_path (str): Where to save the single metrics CSV file.
        :param should_log_metrics (bool): Whether to log metrics at all. Should probably always be True.
        :param hparams_path (str): Where to save the single hyperparameter YAML file.
        :param should_log_hparams (bool): Whether to log hyperparameters to a file.
            Should be False if you want to work with hyperparameters in a dependency file,
            rather than specifying them using command line arguments.
        :param should_make_dirs (bool): If true, the directory structure required by metrics_path
            and hparams_path will be created. Has no effect if the directory structure already exists.
        """
        super(DAGsHubLogger, self).__init__()
        self.logger = LoggerImpl(metrics_path=metrics_path, should_log_metrics=should_log_metrics,
                                 hparams_path=hparams_path, should_log_hparams=should_log_hparams,
                                 should_make_dirs=should_make_dirs, eager_logging=True)

    def on_train_begin(self, logs={}):
        params = {}
        with ignore_exceptions():
            params.update(self.params)
        with ignore_exceptions():
            params['optimizer'] = self.model.optimizer.get_config()
        with ignore_exceptions():
            params['loss'] = self.model.loss.get_config()
        self.logger.log_hyperparams(params)
        self.logger.log_hyperparams(success=False)
        self._epoch = -1
        self._step = 0

    def on_epoch_begin(self, epoch, logs={}):
        self._epoch = epoch

    def on_train_batch_end(self, batch, logs={}):
        self._step += 1
        self.on_epoch_end(None, logs)

    def on_epoch_end(self, epoch, logs={}):
        # At the end of an epoch, logs has more metrics
        metrics = {'epoch': self._epoch, **logs}
        self.logger.log_metrics(metrics, step_num=self._step)

    def on_train_end(self, logs={}):
        self.logger.log_hyperparams(success=True)
        self.logger.save()
        self.logger.close()



import json
from typing import List, Optional, Union

from dagshub.common.analytics import send_analytics_event
from dagshub.data_engine.client.data_client import DataClient
from dagshub.data_engine.client.models import DatasetResult
from dagshub.data_engine.model.datasource import Datasource
from dagshub.data_engine.model.datasource_state import DatasourceState
from dagshub.data_engine.model.errors import DatasetNotFoundError
from dagshub.data_engine.model.query import DatasourceQuery


def get_datasets(repo: str) -> List[Datasource]:
    """
    Gets datasets assigned to the repo.
    Dataset is a combination of a datasource with a query + include/exclude lists
    """
    return _get_datasets(repo)


def get_dataset(repo: str, name: Optional[str] = None, id: Optional[Union[int, str]] = None) -> Datasource:
    """
    Get specific dataset (by name or id)
    """
    assert name is not None or id is not None

    sources = _get_datasets(repo, name, id)
    if len(sources) == 0:
        raise DatasetNotFoundError(repo, id, name)
    return sources[-1]


def _get_datasets(repo: str, name: Optional[str] = None, id: Optional[Union[int, str]] = None) -> List[Datasource]:
    send_analytics_event("Client_DataEngine_getDatasets")
    client = DataClient(repo)
    sources = client.get_datasets(id, name)
    return [_from_gql_result(repo, source) for source in sources]


def _from_gql_result(repo: str, dataset_result: "DatasetResult") -> "Datasource":
    ds = Datasource(DatasourceState.from_gql_result(repo, dataset_result.datasource))

    query_dict = json.loads(dataset_result.datasetQuery)
    if "query" in query_dict:
        ds._query = DatasourceQuery.deserialize(query_dict["query"])

    return ds



import logging

from starlette.requests import Request
from starlette.responses import JSONResponse

from dagshub.common.analytics import send_analytics_event
from dagshub.data_engine.voxel_plugin_server.routes.util import error_handler
from dagshub.data_engine.voxel_plugin_server.utils import get_plugin_state

logger = logging.getLogger(__name__)


@error_handler
async def to_annotate(request: Request):
    plugin_state = get_plugin_state(request)

    ds = plugin_state.datasource
    send_analytics_event("Client_DataEngine_SentToAnnotationWithVoxel", repo=ds.source.repoApi)

    selected = plugin_state.voxel_session.selected_view
    print(selected)
    req_dicts = []
    if selected is None:
        return JSONResponse({"error": "Selection empty"}, status_code=400)
    for sample in selected:
        req_dicts.append({
            "datapoint_id": sample["datapoint_id"],
            "download_url": sample["dagshub_download_url"],
        })
    print(f"Sending to annotation: {req_dicts}")
    # Don't open the project because we're going to open it from the Voxel's plugin code
    link = ds.send_datapoints_to_annotation(req_dicts, open_project=False)
    return JSONResponse({"link": link})



import datetime
import logging
from abc import ABCMeta, abstractmethod
from typing import Any, Dict, Generator, TYPE_CHECKING, Optional

import dateutil.parser
from httpx import Request, Response, Auth

from dagshub.common import config

if TYPE_CHECKING:
    from dagshub.auth.tokens import TokenStorage

logger = logging.getLogger(__name__)


class DagshubAuthenticator(Auth):
    """
    This class contains a token + flow on how to re-init the token in case of failure
    """

    def __init__(self, token: "DagshubTokenABC", token_storage: "TokenStorage", host: str):
        self._token = token
        self._token_storage = token_storage
        self._host = host

    def auth_flow(self, request: Request) -> Generator[Request, Response, None]:
        # TODO: failure mode recovery
        response = yield self._token(request)
        if response.status_code == 401:
            if self.can_renegotiate():
                logger.warning("Request to DagsHub has failed, getting a new token")
                self.renegotiate_token()
                yield self._token(request)
            elif not self._token_storage.is_valid_token(self._token, self._host):
                logger.warning("Your DagsHub token is no longer valid!")

    def can_renegotiate(self):
        # Env var tokens cannot renegotiate, every other token type can
        return not type(self._token) is EnvVarDagshubToken

    def renegotiate_token(self):
        if not self._token_storage.is_valid_token(self._token, self._host):
            self._token_storage.invalidate_token(self._token, self._host)
        self._token = self._token_storage.get_token_object(self._host)

    @property
    def token_text(self) -> str:
        return self._token.token_text

    def __call__(self, request):
        """
        Forward the call to the token
        This is for the codebases that don't rely on httpx
        This one does not have automatic renegotiation!!!

        TODO: need to figure out how to renegotiate in the requests library.
        Probably take a look at how they do NTLM
        """
        return self._token(request)


class TokenDeserializationError(Exception):
    ...


class DagshubTokenABC(metaclass=ABCMeta):
    token_type = "NONE"
    priority = 10000  # Decides which token is given out first to the user

    def __call__(self, request: Request) -> Request:
        request.headers["Authorization"] = f"Bearer {self.token_text}"
        return request

    @staticmethod
    @abstractmethod
    def deserialize(values: Dict[str, Any]):
        ...

    @abstractmethod
    def serialize(self) -> Dict[str, Any]:
        ...

    @property
    @abstractmethod
    def token_text(self) -> str:
        ...

    @property
    @abstractmethod
    def is_expired(self) -> bool:
        ...


class OAuthDagshubToken(DagshubTokenABC):
    token_type = "bearer"
    priority = 1

    def __init__(self, token_value: str, expiry_date: datetime.datetime):
        self.token_value = token_value
        self.expiry_date = expiry_date

    # TODO: override the call function to warn about how much lifetime the token has

    def serialize(self) -> Dict[str, Any]:
        return {
            "access_token": self.token_value,
            "expiry": self.expiry_date.isoformat(),
            "token_type": self.token_type,
        }

    @staticmethod
    def deserialize(values: Dict[str, Any]):
        token_value = values["access_token"]
        expiry_date = values["expiry"]
        try:
            expiry_date = dateutil.parser.parse(expiry_date)
        except dateutil.parser.ParserError as ex:
            raise TokenDeserializationError from ex
        return OAuthDagshubToken(token_value, expiry_date)

    @property
    def token_text(self) -> str:
        return self.token_value

    @property
    def is_expired(self) -> bool:
        return self.expiry_date < datetime.datetime.now(tz=self.expiry_date.tzinfo)

    def __repr__(self):
        return f"Dagshub OAuth token, valid until {self.expiry_date}"


class AppDagshubToken(DagshubTokenABC):
    token_type = "app-token"
    priority = 0

    def __init__(self, token_value: str):
        self.token_value = token_value

    def serialize(self) -> Dict[str, Any]:
        return {
            "access_token": self.token_value,
            "expiry": "never",
            "token_type": self.token_type,
        }

    @staticmethod
    def deserialize(values: Dict[str, Any]):
        return AppDagshubToken(values["access_token"])

    @property
    def token_text(self) -> str:
        return self.token_value

    @property
    def is_expired(self) -> bool:
        return False

    def __repr__(self):
        return "Dagshub App token"


class EnvVarDagshubToken(DagshubTokenABC):
    token_type = "env-var"
    priority = -1

    def __init__(self, token_value: str, host: Optional[str] = None):
        self.token_value = token_value
        self.host = host or config.host

    def serialize(self) -> Dict[str, Any]:
        raise RuntimeError("Can't serialize env var token")

    @staticmethod
    def deserialize(values: Dict[str, Any]):
        raise RuntimeError("Can't deserialize env var token")

    @property
    def token_text(self) -> str:
        return self.token_value

    @property
    def is_expired(self) -> bool:
        return False

    def __repr__(self):
        return f"Dagshub Env Var token for host {self.host}"


class HTTPBearerAuth(Auth):
    """Attaches HTTP Bearer Authorization to the given Request object."""

    def __init__(self, token):
        self.token = token

    def auth_flow(self, request: Request) -> Generator[Request, Response, None]:
        request.headers["Authorization"] = f"Bearer {self.token}"
        yield request

    def __eq__(self, other):
        return all([
            self.token == getattr(other, 'token', None),
        ])

    def __ne__(self, other):
        return not self == other

    def __call__(self, r):
        r.headers['Authorization'] = f'Bearer {self.token}'
        return r



import datetime
import logging
import os
import threading
import traceback
from typing import Optional, Dict, List, Set, Union

import yaml
from httpx import Auth

from dagshub.auth import oauth
from dagshub.auth.token_auth import HTTPBearerAuth, DagshubTokenABC, TokenDeserializationError, AppDagshubToken, \
    EnvVarDagshubToken, DagshubAuthenticator
from dagshub.common import config
from dagshub.common.helpers import http_request
from dagshub.common.util import multi_urljoin

logger = logging.getLogger(__name__)

APP_TOKEN_TYPE = "app-token"


class InvalidTokenError(Exception):
    def __str__(self):
        print("The token is not a valid DagsHub token")


class TokenStorage:
    def __init__(self, cache_location: str = None, **kwargs):
        cache_location = cache_location or config.cache_location
        self.cache_location = cache_location
        self.schema_version = config.TOKENS_CACHE_SCHEMA_VERSION
        self.__token_cache: Optional[Dict[str, List[DagshubTokenABC]]] = None

        # We check tokens only once for validity, so we don't do a lot of redundant requests
        #   maybe there is a point to re-evaluate them once in a while
        self._known_good_tokens: Dict[str, Set[DagshubTokenABC]] = {}

        self.__token_access_lock = threading.RLock()

    @property
    def _token_cache(self):
        if self.__token_cache is None:
            self.__token_cache = self._load_cache_file()
            self.remove_expired_tokens()
        return self.__token_cache

    @property
    def _token_access_lock(self):
        if not hasattr(self, "__token_access_lock"):
            self.__token_access_lock = threading.RLock()
        return self.__token_access_lock

    def remove_expired_tokens(self):
        had_changes = False
        for host, tokens in self._token_cache.items():
            if host == "version":
                continue
            expired_tokens = filter(lambda token: token.is_expired, tokens)
            for t in expired_tokens:
                had_changes = True
                tokens.remove(t)
        if had_changes:
            logger.info("Removed expired tokens from the token cache")
            self._store_cache_file()

    def add_token(self, token: Union[str, DagshubTokenABC], host: str = None, skip_validation=False):
        host = host or config.host

        if type(token) is str:
            token = AppDagshubToken(token)

        if self._token_already_exists(token.token_text, host):
            logger.warning("The added token already exists in the token cache, skipping")
            return

        if not skip_validation:
            if not TokenStorage.is_valid_token(token.token_text, host):
                raise InvalidTokenError

        if host not in self._token_cache:
            self._token_cache[host] = []
        self._token_cache[host].append(token)
        self._store_cache_file()

    def invalidate_token(self, token: DagshubTokenABC, host: str = None):
        host = host or config.host

        try:
            if host in self._token_cache:
                tokens = self._token_cache[host]
                tokens.remove(token)

                if host in self._known_good_tokens:
                    good_token_set = self._known_good_tokens[host]
                    if token in good_token_set:
                        good_token_set.remove(token)

                self._store_cache_file()
        except ValueError:
            logger.warning(f"Token {token} does not exist in the storage")

    def get_authenticator(self, host: str = None, fail_if_no_token: bool = False, **kwargs) -> DagshubAuthenticator:
        """
        Returns the authenticator object, that can renegotiate tokens in case of failure
        """
        host = host or config.host
        token = self.get_token_object(host, fail_if_no_token, **kwargs)
        return DagshubAuthenticator(token, token_storage=self, host=host)

    def get_token_object(self, host: str = None, fail_if_no_token: bool = False, **kwargs) -> DagshubTokenABC:
        """
         This function does following:
         - Iterates over all tokens in the cache for the provided host
         - Finds a first valid token and returns it
         - If it finds an invalid token, it deletes it from the cache

         We're using a set of known good tokens to skip rechecking for token validity every time
         """

        host = host or config.host
        if host == config.host and config.token is not None:
            return EnvVarDagshubToken(config.token, host)

        with self._token_access_lock:
            tokens = self._token_cache.get(host, [])

            if host not in self._known_good_tokens:
                self._known_good_tokens[host] = set()
            good_token_set = self._known_good_tokens[host]
            good_token = None
            token_queue = list(sorted(tokens, key=lambda t: t.priority))

            for token in token_queue:
                if token.is_expired:
                    self.invalidate_token(token, host)
                    good_token_set = self._known_good_tokens[host]

                if token in good_token_set:
                    good_token = token
                    break
                # Check token validity
                elif self.is_valid_token(token, host):
                    good_token = token
                    good_token_set.add(token)
                # Remove invalid token from the cache
                else:
                    self.invalidate_token(token, host)
                    good_token_set = self._known_good_tokens[host]
                if good_token is not None:
                    break

            # Couldn't manage to find a good token after the search
            # Either go through the oauth flow, or throw a runtime error
            if good_token is None:
                if fail_if_no_token:
                    raise RuntimeError(
                        f"No valid tokens found for host '{host}'.\n"
                        "Log into DagsHub by executing `dagshub login` in your terminal")
                else:
                    logger.debug(
                        f"No valid tokens found for host '{host}'. Authenticating with OAuth"
                    )
                    good_token = oauth.oauth_flow(host, **kwargs)
                    tokens.append(good_token)
                    good_token_set.add(good_token)
                    # Save the cache
                    self._token_cache[host] = tokens
                    self._store_cache_file()

            return good_token

    def get_token(self, host: str = None, fail_if_no_token: bool = False, **kwargs) -> str:
        """
        Return the raw token string
        This is a lower level method that cannot do renegotiations, we only return the token itself here.
        Used mainly for setting environment variables, for example for MLflow
        """
        return self.get_token_object(host, fail_if_no_token).token_text

    def _token_already_exists(self, token_text: str, host: str):
        for token in self._token_cache.get(host, []):
            if token.token_text == token_text:
                return True
        return False

    @staticmethod
    def _is_expired(token: Dict[str, str]) -> bool:
        if "expiry" not in token:
            return True
        if token["expiry"] == "never":
            return False
        # Need to cut off the three additional precision numbers in milliseconds, because %f only parses 6 digits
        expiry = token["expiry"][:-4] + "Z"
        expiry_dt = datetime.datetime.strptime(expiry, "%Y-%m-%dT%H:%M:%S.%fZ")
        is_expired = expiry_dt < datetime.datetime.utcnow()
        return is_expired

    @staticmethod
    def is_valid_token(token: Union[str, Auth, DagshubTokenABC], host: str) -> bool:
        """
        Check for token validity

        Args:
            token: token to check validity
            host: which host to connect against
        """
        host = host or config.host
        check_url = multi_urljoin(host, "api/v1/user")
        if type(token) is str:
            auth = HTTPBearerAuth(token)
        else:
            auth = token
        resp = http_request("GET", check_url, auth=auth)

        try:
            # 500's might be ok since they're server errors, so check only for 400's
            assert not (400 <= resp.status_code <= 499)
            if resp.status_code == 200:
                assert "login" in resp.json()
            return True
        except AssertionError:
            return False

    def _load_cache_file(self) -> Dict[str, List[DagshubTokenABC]]:
        logger.debug(f"Loading token cache from {self.cache_location}")
        if not os.path.exists(self.cache_location):
            logger.debug("Token cache file doesn't exist")
            return {}
        try:
            with open(self.cache_location) as f:
                cache_yaml = yaml.load(f, yaml.Loader)
                version = cache_yaml.get("version", "1")
                if version == "1":
                    return self._v1_token_list_parser(cache_yaml)
                raise RuntimeError(f"Don't know how to parse token schema {version}")
        except Exception:
            logger.error(
                f"Error while loading DagsHub token cache: {traceback.format_exc()}"
            )
            raise

    @staticmethod
    def _v1_token_list_parser(cache_yaml: Dict[str, Union[str, List[Dict]]]) -> Dict[str, List[DagshubTokenABC]]:
        res = {}

        token_class_map = {}
        for token_class in DagshubTokenABC.__subclasses__():
            token_class_map[token_class.token_type] = token_class

        for host, tokens in cache_yaml.items():
            if host == "version":
                continue
            if len(tokens) == 0:
                continue
            host_tokens = []
            for token_dict in tokens:
                try:
                    token = token_class_map[token_dict["token_type"]].deserialize(token_dict)
                    host_tokens.append(token)
                except TokenDeserializationError as e:
                    logger.warning(f"Failed to deserialize token {token_dict}: {e}")
            res[host] = host_tokens
        return res

    def _store_cache_file(self):
        logger.debug(f"Dumping token cache to {self.cache_location}")
        try:
            dirpath = os.path.dirname(self.cache_location)
            if not os.path.exists(dirpath):
                os.makedirs(dirpath)
            dict_to_dump = {"version": self.schema_version}
            for host, tokens in self.__token_cache.items():
                dict_to_dump[host] = [t.serialize() for t in tokens]
            with open(self.cache_location, "w") as f:
                yaml.dump(dict_to_dump, f, yaml.Dumper)
        except Exception:
            logger.error(
                f"Error while storing DagsHub token cache: {traceback.format_exc()}"
            )
            raise

    def __getstate__(self):
        d = self.__dict__
        # Don't pickle the lock. This will make it so multiple authenticators might request for tokens at the same time
        # This can lead to e.g. multiple OAuth requests firing at the same time, which is not desirable
        # However, I'm not sure of a good way to solve it
        access_lock_key = f"_{self.__class__.__name__}__token_access_lock"
        if access_lock_key in d:
            del (d[access_lock_key])
        return d

    def __setstate__(self, state):
        self.__dict__ = state


_token_storage: Optional[TokenStorage] = None


def _get_token_storage(**kwargs):
    global _token_storage
    if _token_storage is None:
        _token_storage = TokenStorage(**kwargs)
    return _token_storage


def get_authenticator(**kwargs) -> DagshubAuthenticator:
    """
    Get an authenticator object.
    This object can be used as auth argument for the httpx requests

    The authenticator has renegotiation logic in case where a token gets invalidated
    """
    return _get_token_storage(**kwargs).get_authenticator(**kwargs)


def get_token_object(**kwargs):
    """
    Gets a DagsHub token, by default if no token is found authenticates with OAuth

    Kwargs:
        host (str): URL of a dagshub instance (defaults to dagshub.com)
        cache_location (str): Location of the cache file with the token (defaults to <cache_dir>/dagshub/tokens)
        fail_if_no_token (bool): What to do if token is not found.
            If set to False (default), goes through OAuth flow
            If set to True, throws a RuntimeError
    """
    return _get_token_storage(**kwargs).get_token_object(**kwargs)


def get_token(**kwargs):
    """
    Gets a DagsHub token text, by default if no token is found authenticates with OAuth

    Kwargs:
        host (str): URL of a dagshub instance (defaults to dagshub.com)
        cache_location (str): Location of the cache file with the token (defaults to <cache_dir>/dagshub/tokens)
        fail_if_no_token (bool): What to do if token is not found.
            If set to False (default), goes through OAuth flow
            If set to True, throws a RuntimeError
    """
    return _get_token_storage(**kwargs).get_token(**kwargs)


def add_app_token(token: str, host: Optional[str] = None, **kwargs):
    """
    Adds an application token to the token cache.
    This is a long-lived token that you can add/revoke in your profile settings on DagsHub
    """
    token_obj = AppDagshubToken(token)
    _get_token_storage(**kwargs).add_token(token_obj, host)


def add_oauth_token(host: Optional[str] = None, **kwargs):
    """
    Launches the OAuth flow that generates a short-lived token.
    This will open a new browser window, so this is not a CI/headless friendly function.
    Consider using `add_app_token` or setting the `DAGSHUB_USER_TOKEN` env var in those cases.
    """
    host = host or config.host
    token = oauth.oauth_flow(host)
    _get_token_storage(**kwargs).add_token(token, host, skip_validation=True)



import os
import sys
import tempfile
import shutil

import click
import logging
import git
import zipfile
import tarfile
from http import HTTPStatus
from urllib.parse import urlparse

import dagshub.auth
import dagshub.common.logging_util
from dagshub import init, __version__
from dagshub.common import config, rich_console
from dagshub.upload import create_repo, Repo
from dagshub.common.helpers import http_request, log_message
from dagshub.upload.errors import UpdateNotAllowedError
from dagshub.upload.wrapper import add_dataset_to_repo, DEFAULT_DATA_DIR_NAME


@click.group()
@click.option("--host", default=config.host, help="Hostname of DagsHub instance")
@click.option("-q", "--quiet", is_flag=True, help="Suppress print output")
@click.pass_context
def cli(ctx, host, quiet):
    dagshub.common.logging_util.init_logger()
    ctx.obj = {"host": host.strip("/"), "quiet": quiet or config.quiet}


@cli.command()
@click.argument("project_root", default=".")
@click.option("--repo_url", help="URL of the repo hosted on DagsHub")
@click.option("--branch", help="Repository's branch")
@click.option("-v", "--verbose", default=0, count=True, help="Verbosity level")
@click.option(
    "--debug", default=False, type=bool, help="Run fuse in foreground"
)
@click.option("-q", "--quiet", is_flag=True, help="Suppress print output")
@click.pass_context
def mount(ctx, verbose, quiet, **kwargs):
    """
    Mount a DagsHub Storage folder via FUSE
    """
    # Since pyfuse can crash on init-time, import it here instead of up top
    from dagshub.streaming import mount

    config.quiet = ctx.obj["quiet"] or quiet

    logger = logging.getLogger()
    logger.setLevel(to_log_level(verbose))

    if not kwargs["debug"]:
        # Hide tracebacks of errors, display only error message
        sys.tracebacklimit = 0
    mount(**kwargs)


@cli.group()
@click.pass_context
def setup(ctx):
    """
    Initialize additional functionality in the current repository
    """
    pass


@setup.command("dvc")
@click.option("--repo_name", help="The repository name to set up")
@click.option("--repo_owner", help="Owner of the repository in use (user or organization)")
@click.option("--url", help="DagsHub remote url; either provide --url or repo_name and repo_owner")
@click.option("--host", default=config.DEFAULT_HOST, help="DagsHub instance to which you want to login")
@click.option("-q", "--quiet", is_flag=True, help="Suppress print output")
@click.pass_context
def setup_dvc(ctx, quiet, repo_name, repo_owner, url, host):
    """
    Initialize dvc
    """
    host = host or ctx.obj["host"]
    config.quiet = quiet or ctx.obj["quiet"]
    init(repo_name=repo_name, repo_owner=repo_owner, url=url, root=None, host=host, mlflow=False, dvc=True)


@cli.command()
@click.option("--token", help="Login using a specified token")
@click.option("--host", help="DagsHub instance to which you want to login")
@click.option("-q", "--quiet", is_flag=True, help="Suppress print output")
@click.pass_context
def login(ctx, token, host, quiet):
    """
    Initiate an Oauth authentication process. This process will generate and cache a short-lived token in your
    local machine, to allow you to perform actions that require authentication. After running `dagshub login` you can
    use data streaming and upload files without providing authentication info.
    """
    host = host or ctx.obj["host"]
    config.quiet = quiet or ctx.obj["quiet"]
    if token is not None:
        dagshub.auth.add_app_token(token, host)
        rich_console.print(":white_check_mark: Token added successfully")
    else:
        dagshub.auth.add_oauth_token(host)
        rich_console.print(":white_check_mark: OAuth token added")


def validate_repo(ctx, param, value):
    parts = value.split("/")
    if len(parts) != 2:
        raise click.BadParameter("repo needs to be in the format <repo-owner>/<repo-name>")
    return tuple(parts)


def to_log_level(verbosity):
    if verbosity == 0:
        return logging.WARN
    elif verbosity == 1:
        return logging.INFO
    elif verbosity >= 2:
        return logging.DEBUG


@cli.command()
@click.argument("repo", callback=validate_repo)
@click.argument("filename", type=click.Path(exists=True))
@click.argument("target", required=False)
@click.option("-m", "--message", help="Commit message for the upload")
@click.option("-b", "--branch", help="Branch to upload the file to")
@click.option("--update", is_flag=True, help="Force update existing files/directories")
@click.option("-v", "--verbose", default=0, count=True, help="Verbosity level")
@click.option("-q", "--quiet", is_flag=True, help="Suppress print output")
@click.option("--host", help="DagsHub instance to which you want to login")
@click.option("--versioning", help="Versioning system to be used to upload the file(s)",
              type=click.Choice(["git", "dvc", "auto"]))
@click.pass_context
def upload(ctx,
           filename,
           target,
           repo,
           message,
           branch,
           verbose,
           update,
           quiet,
           host,
           versioning,
           **kwargs):
    """
    Upload FILENAME to REPO at location TARGET.

    FILENAME can be a directory.

    REPO should be of the form <owner>/<repo-name>, i.e nirbarazida/yolov6.

    TARGET should include the full path inside the repo, including the filename itself.
    If TARGET is omitted, it defaults to using the relative path to FILENAME from current working directory,
    or the filename itself if it's not relative to the current working directory.
    """
    config.host = host or ctx.obj["host"] or config.host
    config.quiet = quiet or ctx.obj["quiet"]

    logger = logging.getLogger()
    logger.setLevel(to_log_level(verbose))

    owner, repo_name = repo
    repo = Repo(owner=owner, name=repo_name, branch=branch)
    try:
        repo.upload(local_path=filename, remote_path=target, commit_message=message, force=update,
                    versioning=versioning)
    except UpdateNotAllowedError:
        log_message(":warning: You're trying to update existing files! :warning:\n"
                    "If you want to do that, retry with --update to force the update", logger)


@cli.command()
def version():
    """
    Prints the current installed version of the DagsHub client
    """
    print(__version__)


@cli.group()
def repo():
    """
    Operations on repo: currently only 'create'
    """
    pass


@repo.command()
@click.argument("repo_name")
@click.option("-u", "--upload-data", help="Upload data from specified url to new repository")
@click.option("-c", "--clone", is_flag=True, help="Clone repository locally")
@click.option("-v", "--verbose", default=0, count=True, help="Verbosity level")
@click.option("-q", "--quiet", is_flag=True, help="Suppress print output")
@click.pass_context
def create(ctx,
           repo_name,
           upload_data,
           clone,
           verbose,
           quiet):
    """
    create a repo and:\n
    optional- upload files to 'data' dir,
     .zip and .tar files are extracted, other formats copied as is.
    optional- clone repo locally.\n
    example 1:  dagshub repo create mytutorial -u "http://example.com/data.csv" --clone\n
    example 2:  dagshub --host "https://www.dagshub.com"
                    repo create mytutorial2 -u "http://0.0.0.0:8080/index.html" --clone --verbose
    """

    config.quiet = quiet or ctx.obj["quiet"]

    logger = logging.getLogger()
    logger.setLevel(to_log_level(verbose))

    with tempfile.TemporaryDirectory() as tmp_dir:

        # override default host if provided by --host
        host = ctx.obj["host"]

        # create remote repo
        repo = create_repo(repo_name, host=host)
        log_message(f"Created repo: {host}/{repo.owner}/{repo.name}.git", logger)

        if upload_data:
            # get the data
            res = http_request("GET", upload_data)
            if res.status_code != HTTPStatus.OK:
                raise RuntimeError(f"Could not get file from source (response: {res.status_code}), repo created")

            downloaded_file_name = os.path.basename(urlparse(upload_data).path)

            # save to disk
            with open(downloaded_file_name, 'wb') as fh:
                fh.write(res.content)

            log_message(f"Downloaded and saved {downloaded_file_name}", logger)

            # extract to data dir or move there
            if zipfile.is_zipfile(downloaded_file_name):
                with zipfile.ZipFile(downloaded_file_name, 'r') as zip_ref:
                    zip_ref.extractall(tmp_dir)
            elif tarfile.is_tarfile(downloaded_file_name):
                with tarfile.TarFile(downloaded_file_name, 'r') as tar_ref:
                    tar_ref.extractall(tmp_dir)
            else:
                shutil.move(downloaded_file_name, f"{tmp_dir}/{downloaded_file_name}")

            # upload data dir as DVC to repo
            add_dataset_to_repo(repo, tmp_dir, DEFAULT_DATA_DIR_NAME)
            log_message("Data uploaded to repo", logger)

        if clone:
            # make local repo
            git.Git(repo.name).clone(f"{host}/{repo.owner}/{repo.name}.git")
            log_message(f"Cloned repo to folder {repo.name}", logger)

            # move the data to it,
            # now the local repo resembles the remote but with copy of data
            if upload_data:
                shutil.move(tmp_dir, f"{repo.name}/{DEFAULT_DATA_DIR_NAME}")
                log_message(f"files moved to {repo.name}/{DEFAULT_DATA_DIR_NAME}", logger)

    # clean tmp file/dir if exists
    if upload_data and os.path.exists(downloaded_file_name):
        os.remove(downloaded_file_name)


if __name__ == "__main__":
    cli()



from multiprocessing import Process
from typing import TYPE_CHECKING, Union
from dagshub.common.util import lazy_load
from dagshub.data_engine.client.loaders.base import DagsHubDataset

torch = lazy_load("torch")
torchaudio = lazy_load("torchaudio")
torchvision = lazy_load("torchvision")

if TYPE_CHECKING:
    import torch


class PyTorchDataset(DagsHubDataset, torch.utils.data.Dataset):
    def __init__(self, *args, **kwargs):
        self.tensorlib = Tensorizers
        super().__init__(*args, **kwargs)
        self.type = "torch"


class PyTorchDataLoader(torch.utils.data.DataLoader):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.dataset.order = list(self.sampler)
        if self.dataset.strategy == "background":
            Process(target=self.dataset.pull).start()


class Tensorizers:
    @staticmethod
    def image(filepath: str) -> torch.Tensor:
        return torchvision.io.read_image(filepath).type(torch.float)

    @staticmethod
    def audio(filepath: str) -> torch.Tensor:
        return torchaudio.load(filepath).type(torch.float)

    @staticmethod
    def video(filepath: str) -> torch.Tensor:
        return torchvision.io.read_video(filepath).type(torch.float)

    @staticmethod
    def numeric(num: Union[float, int]) -> torch.Tensor:
        return torch.tensor(num, dtype=torch.float)



from .tokens import get_token, add_app_token, add_oauth_token, InvalidTokenError, get_authenticator
from .oauth import OauthNonInteractiveShellException

__all__ = [
    get_token.__name__,
    add_app_token.__name__,
    add_oauth_token.__name__,
    OauthNonInteractiveShellException.__name__,
    InvalidTokenError.__name__,
    get_authenticator.__name__,
]



from threading import Thread
from typing import Dict, Any, Optional, Union

from dagshub.auth import get_token
from dagshub.auth.token_auth import HTTPBearerAuth
from dagshub.common import config
from dagshub.common.api.repo import RepoAPI
from dagshub.common.helpers import http_request


def send_analytics_event(event_name: str, repo: Optional[Union[int, "RepoAPI"]] = None, **event_data: Dict[str, Any]):
    if config.disable_analytics:
        return
    if event_data is None:
        event_data = {}
    event_data["event"] = event_name
    if repo is not None:
        if type(repo) is int:
            event_data["repo_id"] = repo
        else:
            event_data["repo_id"] = repo.id
    host = config.host
    token = get_token(host=host)
    t = Thread(target=_send, args=(event_data, host, token), daemon=True)
    t.start()


def _send(event_data: Dict[str, Any], host: str, token: str):
    url = f"{host}/api/internal/trackAnalyticsEvent"

    http_request("POST", url, data=event_data, auth=HTTPBearerAuth(token))



from dataclasses import dataclass
from pathlib import Path
from typing import Optional, Any

try:
    from functools import cached_property
except ImportError:
    from cached_property import cached_property

storage_schemas = ["s3", "gs", "azure"]


@dataclass
class DagshubPath:
    """
    Class for handling any path used inside the virtual filesystem

    Attributes:
        fs (DagsHubFilesystem): Filesystem from which this path is assigned
        absolute_path (Path): Absolute path in the system
        relative_path (Optional[Path]): Path relative to the root of the encapsulating FileSystem.
                                        If None, path is outside the FS
        original_path (Path): Original path as it was accessed by the user
    """
    # TODO: this couples this class hard to the fs, need to decouple later
    fs: Any  # Actual type is DagsHubFilesystem, but imports are wonky
    absolute_path: Optional[Path]
    relative_path: Optional[Path]
    original_path: Optional[Path]

    def __post_init__(self):
        # Handle storage paths - translate s3:/bla-bla to .dagshub/storage/s3/bla-bla
        if self.relative_path is not None:
            str_path = self.relative_path.as_posix()
            for storage_schema in storage_schemas:
                if str_path.startswith(f"{storage_schema}:/"):
                    str_path = str_path[len(storage_schema) + 2:]
                    self.relative_path = Path(".dagshub/storage") / storage_schema / str_path
                    self.absolute_path = self.fs.project_root / self.relative_path

    @cached_property
    def name(self):
        return self.absolute_path.name

    @cached_property
    def is_in_repo(self):
        return self.absolute_path is not None and self.relative_path is not None

    @cached_property
    def is_storage_path(self):
        """
        Is path a storage path (stored in a bucket)
        Those paths are accessible via a path like `.dagshub/storage/s3/bucket/...`
        """
        return self.relative_path.as_posix().startswith(".dagshub/storage")

    @cached_property
    def is_passthrough_path(self):
        """
        Is path a "passthrough" path
        A passthrough path is a path that the FS ignores when trying to look up if the file exists on DagsHub
        This includes:
            - .git and .dvc folders - that prevents accidental access to their caches.
                If you need to read with streaming from a .dvc folder (to read config for example), please pull the repo
            - Any /site-packages/ folder - if you have a venv in your repo, python will try to find packages there.
        """
        str_path = self.relative_path.as_posix()
        if "/site-packages/" in str_path or str_path.endswith("/site-packages"):
            return True
        return str_path.startswith(('.git/', '.dvc/')) or str_path in (".git", ".dvc")

    def __truediv__(self, other):
        return DagshubPath(
            absolute_path=self.absolute_path / other,
            relative_path=self.relative_path / other,
            original_path=self.original_path / other,
            fs=self.fs
        )



import traceback
from functools import wraps

from starlette.responses import Response


def error_handler(route):
    """
    Required in order to not break CORS.
    Starlette's CORS middleware is applied only after its default error handler
    This means that any exception ends up being blocked as a request on voxel's side
    """

    @wraps(route)
    async def wrapped(*args, **kwargs):
        try:
            return await route(*args, **kwargs)
        except:  # noqa
            traceback.print_exc()
            return Response(traceback.format_exc(), status_code=500)

    return wrapped
